"""Implements the :stac-ext:`MGRS Extension <mgrs>`."""

import re
from typing import Any, Dict, FrozenSet, Optional, Pattern, Set, Union, cast

import pystac
from pystac.extensions.base import ExtensionManagementMixin, PropertiesExtension
from pystac.extensions.hooks import ExtensionHooks

SCHEMA_URI: str = "https://stac-extensions.github.io/mgrs/v1.0.0/schema.json"
PREFIX: str = "mgrs:"

# Field names
LATITUDE_BAND_PROP: str = PREFIX + "latitude_band"  # required
GRID_SQUARE_PROP: str = PREFIX + "grid_square"  # required
UTM_ZONE_PROP: str = PREFIX + "utm_zone"

LATITUDE_BANDS: FrozenSet[str] = frozenset(
    {
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "J",
        "K",
        "L",
        "M",
        "N",
        "P",
        "Q",
        "R",
        "S",
        "T",
        "U",
        "V",
        "W",
        "X",
    }
)

UTM_ZONES: FrozenSet[int] = frozenset(
    {
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        31,
        32,
        33,
        34,
        35,
        36,
        37,
        38,
        39,
        40,
        41,
        42,
        43,
        44,
        45,
        46,
        47,
        48,
        49,
        50,
        51,
        52,
        53,
        54,
        55,
        56,
        57,
        58,
        59,
        60,
    }
)

GRID_SQUARE_REGEX: str = (
    r"[ABCDEFGHJKLMNPQRSTUVWXYZ][ABCDEFGHJKLMNPQRSTUV](\d{2}|\d{4}|\d{6}|\d{8}|\d{10})?"
)
GRID_SQUARE_PATTERN: Pattern[str] = re.compile(GRID_SQUARE_REGEX)


def validated_latitude_band(v: str) -> str:
    if not isinstance(v, str):
        raise ValueError("Invalid MGRS latitude band: must be str")
    if v not in LATITUDE_BANDS:
        raise ValueError(f"Invalid MGRS latitude band: {v} is not in {LATITUDE_BANDS}")
    return v


def validated_grid_square(v: str) -> str:
    if not isinstance(v, str):
        raise ValueError("Invalid MGRS grid square identifier: must be str")
    if not GRID_SQUARE_PATTERN.fullmatch(v):
        raise ValueError(
            f"Invalid MGRS grid square identifier: {v}"
            f" does not match the regex {GRID_SQUARE_REGEX}"
        )
    return v


def validated_utm_zone(v: Optional[int]) -> Optional[int]:
    if v is not None and not isinstance(v, int):
        raise ValueError("Invalid MGRS utm zone: must be None or str")
    if v is not None and v not in UTM_ZONES:
        raise ValueError(f"Invalid MGRS UTM zone: {v} is not in {UTM_ZONES}")
    return v


class MgrsExtension(
    PropertiesExtension,
    ExtensionManagementMixin[Union[pystac.Item, pystac.Collection]],
):
    """A concrete implementation of :class:`MgrsExtension` on an :class:`~pystac.Item`
    that extends the properties of the Item to include properties defined in the
    :stac-ext:`MGRS Extension <mgrs>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`MgrsExtension.ext` on an :class:`~pystac.Item` to extend it.

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> proj_ext = MgrsExtension.ext(item)
    """

    item: pystac.Item
    """The :class:`~pystac.Item` being extended."""

    properties: Dict[str, Any]
    """The :class:`~pystac.Item` properties, including extension properties."""

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return "<ItemMgrsExtension Item id={}>".format(self.item.id)

    def apply(
        self,
        latitude_band: str,
        grid_square: str,
        utm_zone: Optional[int] = None,
    ) -> None:
        """Applies MGRS extension properties to the extended Item.

        Args:
            latitude_band : REQUIRED. The latitude band of the Item's centroid.
            grid_square : REQUIRED. MGRS grid square of the Item's centroid.
            utm_zone : The UTM Zone of the Item centroid.
        """
        self.latitude_band = validated_latitude_band(latitude_band)
        self.grid_square = validated_grid_square(grid_square)
        self.utm_zone = validated_utm_zone(utm_zone)

    @property
    def latitude_band(self) -> Optional[str]:
        """Get or sets the latitude band of the datasource."""
        return self._get_property(LATITUDE_BAND_PROP, str)

    @latitude_band.setter
    def latitude_band(self, v: str) -> None:
        self._set_property(
            LATITUDE_BAND_PROP, validated_latitude_band(v), pop_if_none=False
        )

    @property
    def grid_square(self) -> Optional[str]:
        """Get or sets the latitude band of the datasource."""
        return self._get_property(GRID_SQUARE_PROP, str)

    @grid_square.setter
    def grid_square(self, v: str) -> None:
        self._set_property(
            GRID_SQUARE_PROP, validated_grid_square(v), pop_if_none=False
        )

    @property
    def utm_zone(self) -> Optional[int]:
        """Get or sets the latitude band of the datasource."""
        return self._get_property(UTM_ZONE_PROP, int)

    @utm_zone.setter
    def utm_zone(self, v: Optional[int]) -> None:
        self._set_property(UTM_ZONE_PROP, validated_utm_zone(v), pop_if_none=False)

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: pystac.Item, add_if_missing: bool = False) -> "MgrsExtension":
        """Extends the given STAC Object with properties from the :stac-ext:`MGRS
        Extension <mgrs>`.

        This extension can be applied to instances of :class:`~pystac.Item`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Item):
            cls.validate_has_extension(obj, add_if_missing)
            return cast(MgrsExtension, MgrsExtension(obj))
        else:
            raise pystac.ExtensionTypeError(
                f"MGRS Extension does not apply to type '{type(obj).__name__}'"
            )


class MgrsExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids: Set[str] = set()
    stac_object_types = {pystac.STACObjectType.ITEM}


MGRS_EXTENSION_HOOKS: ExtensionHooks = MgrsExtensionHooks()



import re
from datetime import datetime
from typing import Any, Dict, List, Optional

import pystac
from pystac.utils import map_opt, str_to_datetime
from shapely.geometry import Polygon, mapping
from stactools.core.io import ReadHrefModifier
from stactools.core.io.xml import XmlElement

from stactools.sentinel2.constants import COORD_ROUNDING, PRODUCT_METADATA_ASSET_KEY
from stactools.sentinel2.constants import SENTINEL2_PROPERTY_PREFIX as s2_prefix
from stactools.sentinel2.utils import fix_z_values


class ProductMetadataError(Exception):
    pass


class ProductMetadata:
    def __init__(
        self, href, read_href_modifier: Optional[ReadHrefModifier] = None
    ) -> None:
        self.href = href
        self._root = XmlElement.from_file(href, read_href_modifier)

        product_info_node = self._root.find("n1:General_Info/Product_Info")
        if product_info_node is None:
            raise ProductMetadataError(
                f"Cannot find product info node for product metadata at {self.href}"
            )
        self.product_info_node = product_info_node

        datatake_node = self.product_info_node.find("Datatake")
        if datatake_node is None:
            raise ProductMetadataError(
                f"Cannot find Datatake node in product metadata at {self.href}"
            )
        self.datatake_node = datatake_node

        granule_node = self.product_info_node.find(
            "Product_Organisation/Granule_List/Granule"
        )
        if granule_node is None:
            raise ProductMetadataError(
                f"Cannot find granule node in product metadata at {self.href}"
            )
        self.granule_node = granule_node

        reflectance_conversion_node = self._root.find(
            "n1:General_Info/Product_Image_Characteristics/Reflectance_Conversion"
        )
        if reflectance_conversion_node is None:
            raise ProductMetadataError(
                f"Could not find reflectance conversion node in product metadata at {self.href}"
            )
        self.reflectance_conversion_node = reflectance_conversion_node

        qa_node = self._root.find("n1:Quality_Indicators_Info")
        if qa_node is None:
            raise ProductMetadataError(
                f"Could not find QA node in product metadata at {self.href}"
            )
        self.qa_node = qa_node

        # BOA_ADD_OFFSET_VALUES_LIST only exists in processing baseline 04.00 and higher
        boa_add_offset_values_list_node = self._root.find(
            "n1:General_Info/Product_Image_Characteristics/BOA_ADD_OFFSET_VALUES_LIST"
        )
        self.boa_add_offset_values_list_node = boa_add_offset_values_list_node

        def _get_geometries():
            geometric_info = self._root.find("n1:Geometric_Info")
            footprint_text = geometric_info.find_text(
                "Product_Footprint/Product_Footprint/Global_Footprint/EXT_POS_LIST"
            )
            if footprint_text is None:
                ProductMetadataError(
                    f"Cannot parse footprint from product metadata at {self.href}"
                )

            footprint_coords = fix_z_values(footprint_text.split(" "))
            footprint_points = [
                p[::-1]
                for p in list(
                    zip(
                        *[
                            iter(
                                round(coord, COORD_ROUNDING)
                                for coord in footprint_coords
                            )
                        ]
                        * 2
                    )
                )
            ]
            footprint_polygon = Polygon(footprint_points)
            geometry = mapping(footprint_polygon)
            bbox = footprint_polygon.bounds
            return bbox, geometry

        self.bbox, self.geometry = _get_geometries()

    @property
    def scene_id(self) -> str:
        """Returns the string to be used for a STAC Item id.

        Removes the processing number and .SAFE extension
        from the product_id defined below.

        Parsed based on the naming convention found here:
        https://sentinel.esa.int/web/sentinel/user-guides/sentinel-2-msi/naming-convention
        """
        product_id = self.product_id
        # Ensure the product id (PRODUCT_URI) is as expected.
        if not product_id.endswith(".SAFE"):
            raise ValueError(
                "Unexpected value found at "
                f"General_Info/Product_Info: {product_id}. "
                "This was expected to follow the sentinel 2 "
                "naming convention, including "
                "ending in .SAFE"
            )
        id_parts = self.product_id.split("_")

        # Remove .SAFE
        id_parts[-1] = id_parts[-1].replace(".SAFE", "")

        # Remove PDGS Processing Baseline number
        id_parts = [part for part in id_parts if not part.startswith("N")]

        return "_".join(id_parts)

    @property
    def product_id(self) -> str:
        result = self.product_info_node.find_text("PRODUCT_URI")
        if result is None:
            raise ValueError(
                "Cannot determine product ID using product metadata " f"at {self.href}"
            )
        else:
            return result

    @property
    def datetime(self) -> datetime:
        time = self.product_info_node.find_text("PRODUCT_START_TIME")
        if time is None:
            raise ValueError(
                "Cannot determine product start time using product metadata "
                f"at {self.href}"
            )
        else:
            return str_to_datetime(time)

    @property
    def image_media_type(self) -> str:
        if self.granule_node.get_attr("imageFormat") == "GeoTIFF":
            return pystac.MediaType.COG
        else:
            return pystac.MediaType.JPEG2000

    @property
    def image_paths(self) -> List[str]:
        extension = ".tif" if self.image_media_type == pystac.MediaType.COG else ".jp2"

        return [f"{x.text}{extension}" for x in self.granule_node.findall("IMAGE_FILE")]

    @property
    def relative_orbit(self) -> Optional[int]:
        return map_opt(int, self.datatake_node.find_text("SENSING_ORBIT_NUMBER"))

    @property
    def orbit_state(self) -> Optional[str]:
        return self.datatake_node.find_text("SENSING_ORBIT_DIRECTION")

    @property
    def platform(self) -> Optional[str]:
        return self.datatake_node.find_text("SPACECRAFT_NAME")

    @property
    def mgrs_tile(self) -> Optional[str]:
        m = re.search(r"_T(\d{2}[a-zA-Z]{3})_", self.href)
        return None if m is None else m.group(1)

    @property
    def metadata_dict(self) -> Dict[str, Any]:
        result = {
            f"{s2_prefix}:product_uri": self.product_id,
            f"{s2_prefix}:generation_time": self.product_info_node.find_text(
                "GENERATION_TIME"
            ),
            f"{s2_prefix}:processing_baseline": self.product_info_node.find_text(
                "PROCESSING_BASELINE"
            ),
            f"{s2_prefix}:product_type": self.product_info_node.find_text(
                "PRODUCT_TYPE"
            ),
            f"{s2_prefix}:datatake_id": self.datatake_node.get_attr(
                "datatakeIdentifier"
            ),
            f"{s2_prefix}:datatake_type": self.datatake_node.find_text("DATATAKE_TYPE"),
            f"{s2_prefix}:datastrip_id": self.granule_node.get_attr(
                "datastripIdentifier"
            ),
            f"{s2_prefix}:granule_id": self.granule_node.get_attr("granuleIdentifier"),
            f"{s2_prefix}:mgrs_tile": self.mgrs_tile,
            f"{s2_prefix}:reflectance_conversion_factor": map_opt(
                float, self.reflectance_conversion_node.find_text("U")
            ),
        }

        return {k: v for k, v in result.items() if v is not None}

    @property
    def boa_add_offsets(self) -> Dict[str, int]:
        if self.boa_add_offset_values_list_node is not None:
            xs = {
                x.get_attr("band_id"): int(x.text)
                for x in self.boa_add_offset_values_list_node.findall("BOA_ADD_OFFSET")
            }
            return {
                "B01": xs["0"],
                "B02": xs["1"],
                "B03": xs["2"],
                "B04": xs["3"],
                "B05": xs["4"],
                "B06": xs["5"],
                "B07": xs["6"],
                "B08": xs["7"],
                "B8A": xs["8"],
                "B09": xs["9"],
                "B10": xs["10"],
                "B11": xs["11"],
                "B12": xs["12"],
            }
        else:
            return {}

    def create_asset(self):
        asset = pystac.Asset(
            href=self.href, media_type=pystac.MediaType.XML, roles=["metadata"]
        )
        return PRODUCT_METADATA_ASSET_KEY, asset



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Dict, Final, List, Optional, Pattern, Tuple

import pystac
from pystac.utils import map_opt
from stactools.core.io import ReadHrefModifier
from stactools.core.io.xml import XmlElement

from stactools.sentinel2.constants import GRANULE_METADATA_ASSET_KEY
from stactools.sentinel2.constants import SENTINEL2_PROPERTY_PREFIX as s2_prefix

BASELINE_PROCESSING: Final[Pattern[str]] = re.compile(r"_N(\d\d\.\d\d)")


class GranuleMetadataError(Exception):
    pass


class GranuleMetadata:
    def __init__(self, href, read_href_modifier: Optional[ReadHrefModifier] = None):
        self.href = href

        self._root = XmlElement.from_file(href, read_href_modifier)

        tile_id = self._root.find_text("n1:General_Info/TILE_ID")
        if tile_id is None:
            raise GranuleMetadataError(
                f"Cannot find granule tile_id granule metadata at {self.href}"
            )
        self.tile_id = tile_id

        geocoding_node = self._root.find("n1:Geometric_Info/Tile_Geocoding")
        if geocoding_node is None:
            raise GranuleMetadataError(f"Cannot find geocoding node in {self.href}")
        self._geocoding_node = geocoding_node

        tile_angles_node = self._root.find("n1:Geometric_Info/Tile_Angles")
        if tile_angles_node is None:
            raise GranuleMetadataError(f"Cannot find tile angles node in {self.href}")
        self._tile_angles_node = tile_angles_node

        self.viewing_angles = ViewingAngle.from_nodes(
            self._tile_angles_node.findall(
                "Mean_Viewing_Incidence_Angle_List/Mean_Viewing_Incidence_Angle"
            )
        )

        self._image_content_node = self._root.find(
            "n1:Quality_Indicators_Info/Image_Content_QI"
        )

        self.resolution_to_shape: Dict[int, Tuple[int, int]] = {}
        for size_node in self._geocoding_node.findall("Size"):
            res = size_node.get_attr("resolution")
            if res is None:
                raise GranuleMetadataError("Size element does not have resolution.")
            nrows = map_opt(int, size_node.find_text("NROWS"))
            if nrows is None:
                raise GranuleMetadataError(
                    f"Could not get rows from size for resolution {res}"
                )
            ncols = map_opt(int, size_node.find_text("NCOLS"))
            if ncols is None:
                raise GranuleMetadataError(
                    f"Could not get columns from size for resolution {res}"
                )
            self.resolution_to_shape[int(res)] = (nrows, ncols)

    @property
    def epsg(self) -> Optional[int]:
        epsg_str = self._geocoding_node.find_text("HORIZONTAL_CS_CODE")
        if epsg_str is None:
            return None
        else:
            return int(epsg_str.split(":")[1])

    @property
    def proj_bbox(self) -> List[float]:
        """The bbox of the image in the CRS of the image data"""
        nrows, ncols = self.resolution_to_shape[10]
        geoposition = self._geocoding_node.find("Geoposition")
        if geoposition is None:
            raise GranuleMetadataError(f"Cannot find geoposition node in {self.href}")
        ulx = map_opt(float, geoposition.find_text("ULX"))
        if ulx is None:
            raise GranuleMetadataError("Could not get upper left X coordinate")
        uly = map_opt(float, geoposition.find_text("ULY"))
        if uly is None:
            raise GranuleMetadataError("Could not get upper left Y coordinate")

        return [ulx, uly - (10 * nrows), ulx + (10 * ncols), uly]

    @property
    def cloudiness_percentage(self) -> Optional[float]:
        return map_opt(
            float,
            self._image_content_node.find_text("CLOUDY_PIXEL_PERCENTAGE"),
        )

    @property
    def mean_solar_zenith(self) -> Optional[float]:
        return map_opt(
            float,
            self._tile_angles_node.find_text("Mean_Sun_Angle/ZENITH_ANGLE"),
        )

    @property
    def mean_solar_azimuth(self) -> Optional[float]:
        return map_opt(
            float,
            self._tile_angles_node.find_text("Mean_Sun_Angle/AZIMUTH_ANGLE"),
        )

    @property
    def metadata_dict(self):
        properties: Dict[str, Optional[float]] = {
            f"{s2_prefix}:product_type": map_opt(
                float, self._image_content_node.find_text("PRODUCT_TYPE")
            ),
            f"{s2_prefix}:orbit_state": map_opt(
                float, self._image_content_node.find_text("SENSING_ORBIT_DIRECTION")
            ),
            f"{s2_prefix}:datatake_type": map_opt(
                float, self._image_content_node.find_text("DATATAKE_TYPE")
            ),
            f"{s2_prefix}:generation_time": map_opt(
                float, self._image_content_node.find_text("GENERATION_TIME")
            ),
            f"{s2_prefix}:relative_orbit": map_opt(
                float, self._image_content_node.find_text("SENSING_ORBIT_NUMBER")
            ),
            f"{s2_prefix}:reflectance_conversion_factor": map_opt(
                float,
                self._image_content_node.find_text(
                    "BOA_ADD_OFFSET_VALUES_LIST/Reflectance_Conversion/U"
                ),
            ),
            f"{s2_prefix}:degraded_msi_data_percentage": map_opt(
                float,
                self._image_content_node.find_text("DEGRADED_MSI_DATA_PERCENTAGE"),
            ),
            f"{s2_prefix}:nodata_pixel_percentage": map_opt(
                float, self._image_content_node.find_text("NODATA_PIXEL_PERCENTAGE")
            ),
            f"{s2_prefix}:saturated_defective_pixel_percentage": map_opt(
                float,
                self._image_content_node.find_text(
                    "SATURATED_DEFECTIVE_PIXEL_PERCENTAGE"
                ),
            ),
            f"{s2_prefix}:dark_features_percentage": map_opt(
                float, self._image_content_node.find_text("DARK_FEATURES_PERCENTAGE")
            ),
            f"{s2_prefix}:cloud_shadow_percentage": map_opt(
                float, self._image_content_node.find_text("CLOUD_SHADOW_PERCENTAGE")
            ),
            f"{s2_prefix}:vegetation_percentage": map_opt(
                float, self._image_content_node.find_text("VEGETATION_PERCENTAGE")
            ),
            f"{s2_prefix}:not_vegetated_percentage": map_opt(
                float, self._image_content_node.find_text("NOT_VEGETATED_PERCENTAGE")
            ),
            f"{s2_prefix}:water_percentage": map_opt(
                float, self._image_content_node.find_text("WATER_PERCENTAGE")
            ),
            f"{s2_prefix}:unclassified_percentage": map_opt(
                float, self._image_content_node.find_text("UNCLASSIFIED_PERCENTAGE")
            ),
            f"{s2_prefix}:medium_proba_clouds_percentage": map_opt(
                float,
                self._image_content_node.find_text("MEDIUM_PROBA_CLOUDS_PERCENTAGE"),
            ),
            f"{s2_prefix}:high_proba_clouds_percentage": map_opt(
                float,
                self._image_content_node.find_text("HIGH_PROBA_CLOUDS_PERCENTAGE"),
            ),
            f"{s2_prefix}:thin_cirrus_percentage": map_opt(
                float, self._image_content_node.find_text("THIN_CIRRUS_PERCENTAGE")
            ),
            f"{s2_prefix}:snow_ice_percentage": map_opt(
                float, self._image_content_node.find_text("SNOW_ICE_PERCENTAGE")
            ),
        }

        return {k: v for k, v in properties.items() if v is not None}

    @property
    def product_id(self) -> str:
        return self.tile_id

    @property
    def scene_id(self) -> str:
        """Returns the string to be used for a STAC Item id.
        Removes the processing number and .SAFE extension
        from the product_id defined below.
        Parsed based on the naming convention found here:
        https://sentinel.esa.int/web/sentinel/user-guides/sentinel-2-msi/naming-convention
        """
        # Ensure the product id (PRODUCT_URI) is as expected.
        id_parts = self.product_id.split("_")

        # Remove PDGS Processing Baseline number
        id_parts = [part for part in id_parts if not part.startswith("N")]

        return "_".join(id_parts)

    @property
    def platform(self) -> Optional[str]:
        if self.tile_id.startswith("S2A"):
            return "sentinel-2a"
        elif self.tile_id.startswith("S2B"):
            return "sentinel-2b"
        else:
            return None

    @property
    def processing_baseline(self) -> Optional[str]:
        """Returns the string to be used for the baseline_processing property
        Parsed based on the naming convention found here:
        https://sentinel.esa.int/web/sentinel/user-guides/sentinel-2-msi/naming-convention
        """
        mgrs_match = BASELINE_PROCESSING.search(self.product_id)
        if mgrs_match:
            return mgrs_match.group(1)
        return None

    @property
    def pvi_filename(self) -> Optional[str]:
        return self._root.find_text("n1:Quality_Indicators_Info/PVI_FILENAME")

    def create_asset(self):
        asset = pystac.Asset(
            href=self.href, media_type=pystac.MediaType.XML, roles=["metadata"]
        )
        return GRANULE_METADATA_ASSET_KEY, asset


@dataclass
class ViewingAngle:
    azimuth: float
    zenith: float

    @classmethod
    def from_nodes(cls, nodes: List[XmlElement]) -> Dict[str, ViewingAngle]:
        angles = dict()
        for node in nodes:
            band_id_str = node.get_attr("bandId")
            if band_id_str is None:
                raise ValueError("expected band id on viewing angle node")
            else:
                band_id = int(band_id_str)
            if band_id < 8:
                band = f"B0{band_id + 1}"
            elif band_id == 8:
                band = "B8A"
            else:
                band = f"B{band_id:02}"
            zenith = float(
                node.find_text_or_throw(
                    "ZENITH_ANGLE", lambda s: ValueError(f"missing ZENITH_ANGLE: {s}")
                )
            )
            azimuth = float(
                node.find_text_or_throw(
                    "AZIMUTH_ANGLE", lambda s: ValueError(f"missing AZIMUTH_ANGLE: {s}")
                )
            )
            angles[band] = cls(azimuth=azimuth, zenith=zenith)
        return angles



import logging
import os
import shutil
from tempfile import TemporaryDirectory

import pystac
from pystac.utils import make_absolute_href
from stactools.core.utils.convert import cogify
from stactools.core.utils.subprocess import call

logger = logging.getLogger(__name__)

# TODO: Refactor this into general-purpose cogification of item asset
# Currently does not allow the specification of specific assets to cogify.


def create_cogs(item):
    cog_directory = os.path.join(os.path.dirname(item.get_self_href()), "cog")
    if not os.path.exists(cog_directory):
        os.makedirs(cog_directory)

    asset_tuples = list(item.assets.items())
    [
        item.add_asset(*create_cog_asset(key, asset, cog_directory))
        for (key, asset) in asset_tuples
        if is_non_cog_image(asset)
    ]


def create_cog_asset(key, asset, path):
    asset_filename, extension = os.path.splitext(os.path.split(asset.href)[1])
    cog_filename = f"{asset_filename}-COG.tif"
    cog_path = os.path.join(path, cog_filename)

    with TemporaryDirectory() as tmp_dir:
        reprojected_path = os.path.join(tmp_dir, f"reprojected{extension}")
        print(f"reprojecting {asset.href}")
        reproject(asset.href, reprojected_path)
        print(f"cogifying {asset.href}")
        cogify(reprojected_path, cog_path)
        shutil.rmtree(tmp_dir, ignore_errors=True)

    asset = pystac.Asset(
        href=make_absolute_href(cog_path),
        media_type=pystac.MediaType.COG,
        roles=["data"],
        title=f"{asset.title} (COG)",
        properties=asset.properties,
    )
    return f"{key}-cog", asset


def is_non_cog_image(asset):
    return asset.media_type != pystac.MediaType.COG and (
        asset.media_type == pystac.MediaType.GEOTIFF
        or asset.media_type == pystac.MediaType.JPEG2000
    )


def reproject(input_path, output_path):
    call(["gdalwarp", "-t_srs", "epsg:3857", input_path, output_path])



import json
import logging
import os
from typing import Optional

import click
from stactools.core.utils.antimeridian import Strategy

from stactools.sentinel2.constants import DEFAULT_TOLERANCE
from stactools.sentinel2.stac import create_item

logger = logging.getLogger(__name__)


def create_sentinel2_command(cli):
    @cli.group("sentinel2", short_help="Commands for working with sentinel2 data")
    def sentinel2():
        pass

    @sentinel2.command(
        "create-item", short_help="Convert a Sentinel2 L2A granule into a STAC item"
    )
    @click.argument("src")
    @click.argument("dst")
    @click.option(
        "-p",
        "--providers",
        help="Path to JSON file containing array of additional providers",
    )
    @click.option(
        "--tolerance",
        type=float,
        default=DEFAULT_TOLERANCE,
        help="Item geometry simplification tolerance, e.g., 0.0001",
    )
    @click.option(
        "--asset-href-prefix",
        help='Prefix for all Asset hrefs instead of default of the "src" value',
    )
    def create_item_command(
        src: str,
        dst: str,
        providers: Optional[str],
        tolerance: float,
        asset_href_prefix: Optional[str],
    ):
        """Creates a STAC Item for a given Sentinel 2 granule

        SRC is the path to the granule
        DST is directory that a STAC Item JSON file will be created
        in. This will have a filename that matches the ID, which will
        be derived from the Sentinel 2 metadata.
        """
        additional_providers = None
        if providers is not None:
            with open(providers) as f:
                additional_providers = json.load(f)

        antimeridian_strategy = "split"
        strategy = Strategy[antimeridian_strategy.upper()]

        item = create_item(
            granule_href=src,
            additional_providers=additional_providers,
            tolerance=tolerance,
            asset_href_prefix=asset_href_prefix,
            antimeridian_strategy=strategy,
        )

        item_path = os.path.join(dst, "{}.json".format(item.id))
        item.set_self_href(item_path)

        item.save_object()

    return sentinel2



import stactools.core

stactools.core.use_fsspec()


def register_plugin(registry):
    # Register subcommands

    from stactools.sentinel2 import commands

    registry.register_subcommand(commands.create_sentinel2_command)


__version__ = "0.4.1"



from typing import Dict, Final, List

import pystac
from pystac.extensions.eo import Band
from pystac.link import Link
from pystac.provider import ProviderRole

SENTINEL2_PROPERTY_PREFIX = "s2"

SENTINEL_LICENSE: Final[Link] = Link(
    rel="license",
    target="https://sentinel.esa.int/documents/247904/690755/Sentinel_Data_Legal_Notice",
)

SENTINEL_INSTRUMENTS: Final[List[str]] = ["msi"]
SENTINEL_CONSTELLATION: Final[str] = "sentinel-2"

SENTINEL_PROVIDER: Final[pystac.Provider] = pystac.Provider(
    name="ESA",
    roles=[ProviderRole.PRODUCER, ProviderRole.PROCESSOR, ProviderRole.LICENSOR],
    url="https://earth.esa.int/web/guest/home",
)

SAFE_MANIFEST_ASSET_KEY: Final[str] = "safe_manifest"
INSPIRE_METADATA_ASSET_KEY: Final[str] = "inspire_metadata"
PRODUCT_METADATA_ASSET_KEY: Final[str] = "product_metadata"
GRANULE_METADATA_ASSET_KEY: Final[str] = "granule_metadata"
DATASTRIP_METADATA_ASSET_KEY: Final[str] = "datastrip_metadata"
TILEINFO_METADATA_ASSET_KEY: Final[str] = "tileinfo_metadata"

DEFAULT_TOLERANCE: Final[float] = 0.0001
COORD_ROUNDING: Final[int] = 6

SENTINEL_BANDS: Final[Dict[str, Band]] = {
    "coastal": Band.create(
        name="coastal",
        common_name="coastal",
        description="Coastal aerosol (band 1)",
        center_wavelength=0.443,
        full_width_half_max=0.027,
    ),
    "blue": Band.create(
        name="blue",
        common_name="blue",
        description="Blue (band 2)",
        center_wavelength=0.490,
        full_width_half_max=0.098,
    ),
    "green": Band.create(
        name="green",
        common_name="green",
        description="Green (band 3)",
        center_wavelength=0.560,
        full_width_half_max=0.045,
    ),
    "red": Band.create(
        name="red",
        common_name="red",
        description="Red (band 4)",
        center_wavelength=0.665,
        full_width_half_max=0.038,
    ),
    "rededge1": Band.create(
        name="rededge1",
        common_name="rededge",
        description="Red edge 1 (band 5)",
        center_wavelength=0.704,
        full_width_half_max=0.019,
    ),
    "rededge2": Band.create(
        name="rededge2",
        common_name="rededge",
        description="Red edge 2 (band 6)",
        center_wavelength=0.740,
        full_width_half_max=0.018,
    ),
    "rededge3": Band.create(
        name="rededge3",
        common_name="rededge",
        description="Red edge 3 (band 7)",
        center_wavelength=0.783,
        full_width_half_max=0.028,
    ),
    "nir": Band.create(
        name="nir",
        common_name="nir",
        description="NIR 1 (band 8)",
        center_wavelength=0.842,
        full_width_half_max=0.145,
    ),
    "nir08": Band.create(
        name="nir08",
        common_name="nir08",
        description="NIR 2 (band 8A)",
        center_wavelength=0.865,
        full_width_half_max=0.033,
    ),
    "nir09": Band.create(
        name="nir09",
        common_name="nir09",
        description="NIR 3 (band 9)",
        center_wavelength=0.945,
        full_width_half_max=0.026,
    ),
    "cirrus": Band.create(
        name="cirrus",
        common_name="cirrus",
        description="Cirrus (band 10)",
        center_wavelength=1.3735,
        full_width_half_max=0.075,
    ),
    "swir16": Band.create(
        name="swir16",
        common_name="swir16",
        description="SWIR 1 (band 11)",
        center_wavelength=1.610,
        full_width_half_max=0.143,
    ),
    "swir22": Band.create(
        name="swir22",
        common_name="swir22",
        description="SWIR 2 (band 12)",
        center_wavelength=2.190,
        full_width_half_max=0.242,
    ),
}

# A dict describing the resolutions that are
# available for each band as separate assets.
# The first resolution is the sensor gsd; others
# are downscaled versions.
UNSUFFIXED_BAND_RESOLUTION: Final[Dict[str, int]] = {
    "coastal": 60,
    "blue": 10,
    "green": 10,
    "red": 10,
    "rededge1": 20,
    "rededge2": 20,
    "rededge3": 20,
    "nir": 10,
    "nir08": 20,
    "nir09": 60,
    "cirrus": 60,
    "swir16": 20,
    "swir22": 20,
}

BANDS_TO_ASSET_NAME: Final[Dict[str, str]] = {
    "B01": "coastal",
    "B02": "blue",
    "B03": "green",
    "B04": "red",
    "B05": "rededge1",
    "B06": "rededge2",
    "B07": "rededge3",
    "B08": "nir",
    "B8A": "nir08",
    "B09": "nir09",
    "B10": "cirrus",
    "B11": "swir16",
    "B12": "swir22",
}

L2A_IMAGE_PATHS: Final[List[str]] = [
    "preview.jpg",
    "R10m/B04.jp2",
    "R10m/B03.jp2",
    "R10m/B02.jp2",
    "R10m/WVP.jp2",
    "R10m/AOT.jp2",
    "R10m/TCI.jp2",
    "R10m/B08.jp2",
    "R20m/B12.jp2",
    "R20m/B06.jp2",
    "R20m/B07.jp2",
    "R20m/B05.jp2",
    "R20m/B11.jp2",
    "R20m/B04.jp2",
    "R20m/B03.jp2",
    "R20m/B02.jp2",
    "R20m/WVP.jp2",
    "R20m/B8A.jp2",
    "R20m/SCL.jp2",
    "R20m/AOT.jp2",
    "R20m/TCI.jp2",
    "R20m/B08.jp2",
    "R60m/B12.jp2",
    "R60m/B06.jp2",
    "R60m/B07.jp2",
    "R60m/B05.jp2",
    "R60m/B11.jp2",
    "R60m/B04.jp2",
    "R60m/B01.jp2",
    "R60m/B03.jp2",
    "R60m/B02.jp2",
    "R60m/WVP.jp2",
    "R60m/B8A.jp2",
    "R60m/SCL.jp2",
    "R60m/AOT.jp2",
    "R60m/B09.jp2",
    "R60m/TCI.jp2",
    "R60m/B08.jp2",
]

L1C_IMAGE_PATHS: Final[List[str]] = [
    "preview.jpg",
    "B01.jp2",
    "B02.jp2",
    "B03.jp2",
    "B04.jp2",
    "B05.jp2",
    "B06.jp2",
    "B07.jp2",
    "B08.jp2",
    "B8A.jp2",
    "B09.jp2",
    "B10.jp2",
    "B11.jp2",
    "B12.jp2",
    "TCI.jp2",
]



import logging
import os
import re
from dataclasses import dataclass
from datetime import datetime
from itertools import chain
from typing import Any, Dict, Final, List, Optional, Pattern, Tuple

import pystac
import shapely
from pystac.extensions.eo import Band, EOExtension
from pystac.extensions.grid import GridExtension
from pystac.extensions.projection import ProjectionExtension
from pystac.extensions.raster import DataType, RasterBand, RasterExtension
from pystac.extensions.sat import OrbitState, SatExtension
from pystac.extensions.view import ViewExtension
from stactools.core.io import ReadHrefModifier
from stactools.core.projection import reproject_geom, transform_from_bbox
from stactools.core.utils import antimeridian
from stactools.core.utils.antimeridian import Strategy

from stactools.sentinel2.constants import (
    BANDS_TO_ASSET_NAME,
    COORD_ROUNDING,
    DATASTRIP_METADATA_ASSET_KEY,
    DEFAULT_TOLERANCE,
    INSPIRE_METADATA_ASSET_KEY,
    L1C_IMAGE_PATHS,
    L2A_IMAGE_PATHS,
)
from stactools.sentinel2.constants import SENTINEL2_PROPERTY_PREFIX as s2_prefix
from stactools.sentinel2.constants import (
    SENTINEL_BANDS,
    SENTINEL_CONSTELLATION,
    SENTINEL_INSTRUMENTS,
    SENTINEL_LICENSE,
    SENTINEL_PROVIDER,
    UNSUFFIXED_BAND_RESOLUTION,
)
from stactools.sentinel2.granule_metadata import GranuleMetadata, ViewingAngle
from stactools.sentinel2.mgrs import MgrsExtension
from stactools.sentinel2.product_metadata import ProductMetadata
from stactools.sentinel2.safe_manifest import SafeManifest
from stactools.sentinel2.tileinfo_metadata import TileInfoMetadata
from stactools.sentinel2.utils import extract_gsd

logger = logging.getLogger(__name__)

MGRS_PATTERN: Final[Pattern[str]] = re.compile(
    r"_T(\d{1,2})([CDEFGHJKLMNPQRSTUVWX])([ABCDEFGHJKLMNPQRSTUVWXYZ][ABCDEFGHJKLMNPQRSTUV])"
)

TCI_PATTERN: Final[Pattern[str]] = re.compile(r"[_/]TCI[_.]")
AOT_PATTERN: Final[Pattern[str]] = re.compile(r"[_/]AOT[_.]")
WVP_PATTERN: Final[Pattern[str]] = re.compile(r"[_/]WVP[_.]")
SCL_PATTERN: Final[Pattern[str]] = re.compile(r"[_/]SCL[_.]")
THUMBNAIL_PATTERN: Final[Pattern[str]] = re.compile(r"[_/]preview[_.]")

BAND_PATTERN: Final[Pattern[str]] = re.compile(r"[_/](B\w{2})")
IS_TCI_PATTERN: Final[Pattern[str]] = re.compile(r"[_/]TCI")
BAND_ID_PATTERN: Final[Pattern[str]] = re.compile(r"[_/](B\d[A\d])")
RESOLUTION_PATTERN: Final[Pattern[str]] = re.compile(r"(\w{2}m)")

RGB_BANDS: Final[List[Band]] = [
    SENTINEL_BANDS["red"],
    SENTINEL_BANDS["green"],
    SENTINEL_BANDS["blue"],
]

DEFAULT_SCALE = 0.0001


@dataclass(frozen=True)
class Metadata:
    scene_id: str
    cloudiness_percentage: Optional[float]
    extra_assets: Dict[str, pystac.Asset]
    geometry: Dict[str, Any]
    bbox: List[float]
    datetime: datetime
    platform: str
    metadata_dict: Dict[str, Any]
    image_media_type: str
    image_paths: List[str]
    epsg: int
    proj_bbox: List[float]
    resolution_to_shape: Dict[int, Tuple[int, int]]
    processing_baseline: str
    viewing_angles: Dict[str, ViewingAngle]
    orbit_state: Optional[str] = None
    relative_orbit: Optional[int] = None
    sun_azimuth: Optional[float] = None
    sun_zenith: Optional[float] = None
    boa_add_offsets: Optional[Dict[str, int]] = None


def create_item(
    granule_href: str,
    tolerance: float = DEFAULT_TOLERANCE,
    additional_providers: Optional[List[pystac.Provider]] = None,
    read_href_modifier: Optional[ReadHrefModifier] = None,
    asset_href_prefix: Optional[str] = None,
    antimeridian_strategy: Strategy = Strategy.SPLIT,
) -> pystac.Item:
    """Create a STC Item from a Sentinel 2 granule.

    Arguments:
        granule_href: The HREF to the granule. This is expected to be a path
            to a SAFE archive, e.g. https://sentinel2l2a01.blob.core.windows.net/sentinel2-l2/01/C/CV/2016/03/27/S2A_MSIL2A_20160327T204522_N0212_R128_T01CCV_20210214T042702.SAFE,
            or a partial S3 object path, e.g. s3://sentinel-s2-l2a/tiles/10/S/DG/2018/12/31/0/
        tolerance: Determines the level of simplification of the geometry
        additional_providers: Optional list of additional providers to set into the Item
        read_href_modifier: A function that takes an HREF and returns a modified HREF.
            This can be used to modify a HREF to make it readable, e.g. appending
            an Azure SAS token or creating a signed URL.
        asset_href_prefix: The URL prefix to apply to the asset hrefs
        antimeridian_strategy (Antimeridian): Either split on -180 or
            normalize geometries so all longitudes are either positive or
            negative.

    Returns:
        pystac.Item: An item representing the Sentinel 2 scene
    """  # noqa

    if granule_href.lower().endswith(".safe"):
        metadata = metadata_from_safe_manifest(granule_href, read_href_modifier)
    else:
        metadata = metadata_from_granule_metadata(
            granule_href, read_href_modifier, tolerance
        )
    created = datetime.now().strftime("%Y-%m-%dT%H:%MZ")

    item = pystac.Item(
        id=metadata.scene_id,
        geometry=metadata.geometry,
        bbox=metadata.bbox,
        datetime=metadata.datetime,
        properties={"created": created},
    )

    # Handle antimeridian if necessary
    antimeridian.fix_item(item, antimeridian_strategy)

    # --Common metadata--

    item.common_metadata.providers = [SENTINEL_PROVIDER]

    if additional_providers is not None:
        item.common_metadata.providers.extend(additional_providers)

    item.common_metadata.platform = metadata.platform.lower()
    item.common_metadata.constellation = SENTINEL_CONSTELLATION
    item.common_metadata.instruments = SENTINEL_INSTRUMENTS

    # --Extensions--

    # eo
    eo = EOExtension.ext(item, add_if_missing=True)
    eo.cloud_cover = metadata.cloudiness_percentage
    RasterExtension.add_to(item)

    # sat
    if metadata.orbit_state or metadata.relative_orbit:
        sat = SatExtension.ext(item, add_if_missing=True)
        sat.orbit_state = (
            OrbitState(metadata.orbit_state.lower()) if metadata.orbit_state else None
        )
        sat.relative_orbit = metadata.relative_orbit

    # Projection Extension
    projection = ProjectionExtension.ext(item, add_if_missing=True)
    projection.epsg = metadata.epsg
    if projection.epsg is None:
        raise ValueError(
            f"Could not determine EPSG code for {granule_href}; which is required."
        )
    centroid = shapely.geometry.shape(item.geometry).centroid
    projection.centroid = {"lat": round(centroid.y, 5), "lon": round(centroid.x, 5)}

    # MGRS and Grid Extension
    mgrs_match = MGRS_PATTERN.search(metadata.scene_id)
    if mgrs_match and len(mgrs_groups := mgrs_match.groups()) == 3:
        mgrs = MgrsExtension.ext(item, add_if_missing=True)
        mgrs.utm_zone = int(mgrs_groups[0])
        mgrs.latitude_band = mgrs_groups[1]
        mgrs.grid_square = mgrs_groups[2]
        grid = GridExtension.ext(item, add_if_missing=True)
        grid.code = f"MGRS-{mgrs.utm_zone}{mgrs.latitude_band}{mgrs.grid_square}"
    else:
        logger.error(
            f"Error populating MGRS and Grid Extensions fields from ID: {metadata.scene_id}"
        )

    # View Extension
    view = ViewExtension.ext(item, add_if_missing=True)
    view.sun_azimuth = metadata.sun_azimuth
    if msz := metadata.sun_zenith:
        view.sun_elevation = 90 - msz

    # s2 properties
    item.properties.update(metadata.metadata_dict)

    # --Assets--

    image_assets = dict(
        [
            image_asset_from_href(
                asset_href=os.path.join(asset_href_prefix or granule_href, image_path),
                resolution_to_shape=metadata.resolution_to_shape,
                proj_bbox=metadata.proj_bbox,
                media_type=metadata.image_media_type,
                processing_baseline=metadata.processing_baseline,
                viewing_angles=metadata.viewing_angles,
                boa_add_offsets=metadata.boa_add_offsets,
            )
            for image_path in metadata.image_paths
        ]
    )

    for key, asset in chain(image_assets.items(), metadata.extra_assets.items()):
        assert key not in item.assets
        item.add_asset(key, asset)

    # --Links--

    item.links.append(SENTINEL_LICENSE)

    return item


def image_asset_from_href(
    asset_href: str,
    resolution_to_shape: Dict[int, Tuple[int, int]],
    proj_bbox: List[float],
    media_type: Optional[str],
    processing_baseline: str,
    viewing_angles: Dict[str, ViewingAngle],
    boa_add_offsets: Optional[Dict[str, int]] = None,
) -> Tuple[str, pystac.Asset]:
    logger.debug(f"Creating asset for image {asset_href}")

    _, ext = os.path.splitext(asset_href)
    if media_type is not None:
        asset_media_type = media_type
    else:
        if ext.lower() == ".jp2":
            asset_media_type = pystac.MediaType.JPEG2000
        elif ext.lower() in [".tiff", ".tif"]:
            asset_media_type = pystac.MediaType.GEOTIFF
        else:
            raise Exception(f"Must supply a media type for asset : {asset_href}")

    # Handle preview image

    if "_PVI" in asset_href:
        asset = pystac.Asset(
            href=asset_href,
            media_type=asset_media_type,
            title="True color preview",
            roles=["data", "reflectance"],
        )
        asset_eo = EOExtension.ext(asset)
        asset_eo.bands = RGB_BANDS
        return "preview", asset
    elif THUMBNAIL_PATTERN.search(asset_href):
        # thumbnail image
        asset = pystac.Asset(
            href=asset_href,
            media_type=pystac.MediaType.JPEG,
            title="Thumbnail image",
            roles=["thumbnail"],
        )
        return "thumbnail", asset

    # Extract gsd and proj info
    resolution = extract_gsd(asset_href)
    if resolution is None:
        # in Level-1C we can deduct the spatial resolution from the band ID or
        # asset name
        band_id_search = BAND_PATTERN.search(asset_href)
        if band_id_search:
            resolution = highest_asset_res(band_id_search.group(1))
        elif IS_TCI_PATTERN.search(asset_href):
            resolution = 10

    shape = list(resolution_to_shape[int(resolution)])
    transform = transform_from_bbox(proj_bbox, shape)

    def set_asset_properties(_asset: pystac.Asset, _band_gsd: Optional[int] = None):
        if _band_gsd:
            pystac.CommonMetadata(_asset).gsd = _band_gsd
        asset_projection = ProjectionExtension.ext(_asset)
        asset_projection.shape = shape
        asset_projection.bbox = proj_bbox
        asset_projection.transform = transform

    # Handle band image

    band_id_search = BAND_ID_PATTERN.search(asset_href)
    if band_id_search:
        try:
            band_id = band_id_search.group(1)
            asset_res = resolution
            band = band_from_band_id(band_id)
        except KeyError:
            # Level-1C have different names
            band_id = os.path.splitext(asset_href)[0].split("_")[-1]
            band = band_from_band_id(band_id)
            asset_res = highest_asset_res(band_id_search.group(1))

        # Get the asset resolution from the file name.
        # If the asset resolution is the band GSD, then
        # include the gsd information for that asset. Otherwise,
        # do not include the GSD information in the asset
        # as this may be confusing for users given that the
        # raster spatial resolution and gsd will differ.
        # See https://github.com/radiantearth/stac-spec/issues/1096
        band_gsd: Optional[int] = None
        if asset_res == highest_asset_res(band_id):
            asset_id = BANDS_TO_ASSET_NAME[band_id]
            band_gsd = asset_res
        else:
            # If this isn't the default resolution, use the raster
            # resolution in the asset key.
            # TODO: Use the raster extension and spatial_resolution
            # property to encode the spatial resolution of all assets.
            asset_id = f"{BANDS_TO_ASSET_NAME[band_id]}_{int(asset_res)}m"

        asset = pystac.Asset(
            href=asset_href,
            media_type=asset_media_type,
            title=f"{band.description} - {asset_res}m",
            roles=["data", "reflectance"],
        )
        viewing_angle = viewing_angles[band_id]
        # We can't use the ViewExtension here until
        # https://github.com/stac-utils/pystac/issues/793 is fixed
        asset.extra_fields["view:azimuth"] = viewing_angle.azimuth
        asset.extra_fields["view:incidence_angle"] = viewing_angle.zenith

        asset_eo = EOExtension.ext(asset)
        asset_eo.bands = [band_from_band_id(band_id)]
        set_asset_properties(asset, band_gsd)

        RasterExtension.ext(asset).bands = raster_bands(
            boa_add_offsets, processing_baseline, band_id, resolution
        )

    # Handle auxiliary images
    elif TCI_PATTERN.search(asset_href):
        # True color
        asset = pystac.Asset(
            href=asset_href,
            media_type=asset_media_type,
            title="True color image",
            roles=["visual"],
        )
        asset_eo = EOExtension.ext(asset)
        asset_eo.bands = RGB_BANDS
        set_asset_properties(asset)

        maybe_res = extract_gsd(asset_href)
        asset_id = f"visual_{maybe_res}m" if maybe_res and maybe_res != 10 else "visual"
    elif AOT_PATTERN.search(asset_href):
        # Aerosol
        asset = pystac.Asset(
            href=asset_href,
            media_type=asset_media_type,
            title="Aerosol optical thickness (AOT)",
            roles=["data", "reflectance"],
        )
        set_asset_properties(asset)

        RasterExtension.ext(asset).bands = [
            RasterBand.create(
                nodata=0,
                spatial_resolution=resolution,
                data_type=DataType.UINT16,
                bits_per_sample=15,
                scale=0.001,
                offset=0,
            )
        ]

        maybe_res = extract_gsd(asset_href)
        asset_id = mk_asset_id(maybe_res, "aot")
    elif WVP_PATTERN.search(asset_href):
        # Water vapor
        asset = pystac.Asset(
            href=asset_href,
            media_type=asset_media_type,
            title="Water vapour (WVP)",
            roles=["data", "reflectance"],
        )
        set_asset_properties(asset)

        RasterExtension.ext(asset).bands = [
            RasterBand.create(
                nodata=0,
                spatial_resolution=resolution,
                data_type=DataType.UINT16,
                bits_per_sample=15,
                unit="cm",
                scale=0.001,
                offset=0,
            )
        ]

        maybe_res = extract_gsd(asset_href)
        asset_id = mk_asset_id(maybe_res, "wvp")
    elif SCL_PATTERN.search(asset_href):
        # Classification map
        asset = pystac.Asset(
            href=asset_href,
            media_type=asset_media_type,
            title="Scene classification map (SCL)",
            roles=["data", "reflectance"],
        )
        set_asset_properties(asset)

        RasterExtension.ext(asset).bands = [
            RasterBand.create(
                nodata=0,
                spatial_resolution=resolution,
                data_type=DataType.UINT8,
            )
        ]

        maybe_res = extract_gsd(asset_href)
        asset_id = mk_asset_id(maybe_res, "scl")
    else:
        raise ValueError(f"Unexpected asset: {asset_href}")

    return asset_id, asset


def band_from_band_id(band_id):
    return SENTINEL_BANDS[BANDS_TO_ASSET_NAME[band_id]]


def highest_asset_res(band_id: str) -> int:
    return UNSUFFIXED_BAND_RESOLUTION[BANDS_TO_ASSET_NAME[band_id]]


def mk_asset_id(maybe_res: Optional[int], name: str):
    return f"{name.lower()}_{maybe_res}m" if maybe_res and maybe_res != 20 else name


# this is used for SAFE archive format
def metadata_from_safe_manifest(
    granule_href: str, read_href_modifier: Optional[ReadHrefModifier]
) -> Metadata:
    safe_manifest = SafeManifest(granule_href, read_href_modifier)
    product_metadata = ProductMetadata(
        safe_manifest.product_metadata_href, read_href_modifier
    )
    granule_metadata = GranuleMetadata(
        safe_manifest.granule_metadata_href, read_href_modifier
    )
    extra_assets = dict(
        [
            safe_manifest.create_asset(),
            product_metadata.create_asset(),
            granule_metadata.create_asset(),
            (
                INSPIRE_METADATA_ASSET_KEY,
                pystac.Asset(
                    href=safe_manifest.inspire_metadata_href,
                    media_type=pystac.MediaType.XML,
                    roles=["metadata"],
                ),
            ),
            (
                DATASTRIP_METADATA_ASSET_KEY,
                pystac.Asset(
                    href=safe_manifest.datastrip_metadata_href,
                    media_type=pystac.MediaType.XML,
                    roles=["metadata"],
                ),
            ),
        ]
    )

    if granule_metadata.pvi_filename is not None:
        extra_assets["preview"] = pystac.Asset(
            href=os.path.join(granule_href, granule_metadata.pvi_filename),
            media_type=product_metadata.image_media_type,
            roles=["thumbnail"],
        )

    return Metadata(
        scene_id=product_metadata.scene_id,
        extra_assets=extra_assets,
        geometry=product_metadata.geometry,
        bbox=[round(v, COORD_ROUNDING) for v in product_metadata.bbox],
        datetime=product_metadata.datetime,
        platform=product_metadata.platform,
        orbit_state=product_metadata.orbit_state,
        relative_orbit=product_metadata.relative_orbit,
        metadata_dict={
            **product_metadata.metadata_dict,
            **granule_metadata.metadata_dict,
        },
        image_media_type=product_metadata.image_media_type,
        image_paths=product_metadata.image_paths,
        cloudiness_percentage=granule_metadata.cloudiness_percentage,
        epsg=granule_metadata.epsg,
        proj_bbox=[round(v, COORD_ROUNDING) for v in granule_metadata.proj_bbox],
        resolution_to_shape=granule_metadata.resolution_to_shape,
        sun_zenith=granule_metadata.mean_solar_zenith,
        sun_azimuth=granule_metadata.mean_solar_azimuth,
        processing_baseline=granule_metadata.processing_baseline,
        boa_add_offsets=product_metadata.boa_add_offsets,
        viewing_angles=granule_metadata.viewing_angles,
    )


# this is used for the Sinergise S3 format,
# e.g., s3://sentinel-s2-l1c/tiles/10/S/DG/2018/12/31/0/
def metadata_from_granule_metadata(
    granule_metadata_href: str,
    read_href_modifier: Optional[ReadHrefModifier],
    tolerance: float,
) -> Metadata:
    granule_metadata = GranuleMetadata(
        os.path.join(granule_metadata_href, "metadata.xml"), read_href_modifier
    )
    tileinfo_metadata = TileInfoMetadata(
        os.path.join(granule_metadata_href, "tileInfo.json"), read_href_modifier
    )

    product_metadata = None
    if os.path.exists(f := os.path.join(granule_metadata_href, "product_metadata.xml")):
        product_metadata = ProductMetadata(f, read_href_modifier)
    elif granule_metadata_href.startswith("https://roda.sentinel-hub.com"):
        f = (
            granule_metadata_href.split("tiles/")[0]
            + tileinfo_metadata.product_path
            + "/metadata.xml"
        )
        product_metadata = ProductMetadata(f, read_href_modifier)

    geometry = shapely.geometry.shape(
        reproject_geom(
            f"epsg:{granule_metadata.epsg}", "epsg:4326", tileinfo_metadata.geometry
        )
    ).simplify(tolerance)

    extra_assets = dict(
        [
            granule_metadata.create_asset(),
            tileinfo_metadata.create_asset(),
        ]
    )

    image_paths = (
        L2A_IMAGE_PATHS if "_L2A_" in granule_metadata.scene_id else L1C_IMAGE_PATHS
    )

    metadata_dict = {
        **granule_metadata.metadata_dict,
        **tileinfo_metadata.metadata_dict,
        f"{s2_prefix}:processing_baseline": granule_metadata.processing_baseline,
    }
    if product_metadata is not None:
        metadata_dict.update(**product_metadata.metadata_dict)

    return Metadata(
        scene_id=granule_metadata.scene_id,
        extra_assets=extra_assets,
        metadata_dict=metadata_dict,
        cloudiness_percentage=granule_metadata.cloudiness_percentage,
        epsg=granule_metadata.epsg,
        proj_bbox=granule_metadata.proj_bbox,
        resolution_to_shape=granule_metadata.resolution_to_shape,
        geometry=shapely.geometry.mapping(geometry),
        bbox=geometry.bounds,
        datetime=tileinfo_metadata.datetime,
        platform=granule_metadata.platform,
        image_media_type=pystac.MediaType.JPEG2000,
        image_paths=image_paths,
        sun_zenith=granule_metadata.mean_solar_zenith,
        sun_azimuth=granule_metadata.mean_solar_azimuth,
        processing_baseline=granule_metadata.processing_baseline,
        boa_add_offsets=product_metadata.boa_add_offsets if product_metadata else None,
        viewing_angles=granule_metadata.viewing_angles,
    )


def offset_for_pb(processing_baseline: str) -> float:
    if processing_baseline < "04.00":
        return 0
    else:
        return -0.1


def raster_bands(
    boa_add_offsets: Optional[Dict[str, int]],
    processing_baseline: str,
    band_id: str,
    resolution: float,
) -> List[RasterBand]:
    # prior to processing baseline 04.00, scale and offset were
    # defined out of band, so handle that case
    offset = (
        round(boa_add_offsets[band_id] * DEFAULT_SCALE, 6)
        if boa_add_offsets
        else offset_for_pb(processing_baseline)
    )

    return [
        RasterBand.create(
            nodata=0,
            spatial_resolution=resolution,
            data_type=DataType.UINT16,
            bits_per_sample=15,
            scale=DEFAULT_SCALE,
            offset=offset,
        )
    ]



import os
from typing import List, Optional, Tuple

import pystac
from stactools.core.io import ReadHrefModifier
from stactools.core.io.xml import XmlElement

from stactools.sentinel2.constants import SAFE_MANIFEST_ASSET_KEY


class ManifestError(Exception):
    pass


class SafeManifest:
    def __init__(
        self, granule_href: str, read_href_modifier: Optional[ReadHrefModifier] = None
    ):
        self.granule_href = granule_href
        self.href = os.path.join(granule_href, "manifest.safe")

        root = XmlElement.from_file(self.href, read_href_modifier)
        self._data_object_section = root.find("dataObjectSection")
        if self._data_object_section is None:
            raise ManifestError(
                f"Manifest at {self.href} does not have a dataObjectSection"
            )

    def _find_href(self, xpaths: List[str]) -> Optional[str]:
        file_path = None
        for xpath in xpaths:
            file_path = self._data_object_section.find_attr("href", xpath)
            if file_path is not None:
                break

        if file_path is None:
            return None
        else:
            # Remove relative prefix that some paths have
            file_path = file_path.strip("./")
            return os.path.join(self.granule_href, file_path)

    @property
    def product_metadata_href(self) -> Optional[str]:
        return self._find_href(
            [
                'dataObject[@ID="S2_Level-1C_Product_Metadata"]/byteStream/fileLocation',
                'dataObject[@ID="S2_Level-2A_Product_Metadata"]/byteStream/fileLocation',
            ]
        )

    @property
    def inspire_metadata_href(self) -> Optional[str]:
        return self._find_href(
            ['dataObject[@ID="INSPIRE_Metadata"]/byteStream/fileLocation']
        )

    @property
    def datastrip_metadata_href(self) -> Optional[str]:
        return self._find_href(
            [
                'dataObject[@ID="S2_Level-1C_Datastrip1_Metadata"]/byteStream/fileLocation',
                'dataObject[@ID="S2_Level-2A_Datastrip1_Metadata"]/byteStream/fileLocation',
            ]
        )

    @property
    def granule_metadata_href(self) -> Optional[str]:
        return self._find_href(
            [
                'dataObject[@ID="S2_Level-2A_Tile1_Data"]/byteStream/fileLocation',
                'dataObject[@ID="S2_Level-1C_Tile1_Metadata"]/byteStream/fileLocation',
                'dataObject[@ID="S2_Level-2A_Tile1_Metadata"]/byteStream/fileLocation',
            ]
        )

    def create_asset(self) -> Tuple[str, pystac.Asset]:
        asset = pystac.Asset(
            href=self.href, media_type=pystac.MediaType.XML, roles=["metadata"]
        )
        return SAFE_MANIFEST_ASSET_KEY, asset



import re
from typing import Final, List, Optional, Pattern

import shapely
from pystac import Item
from shapely.geometry import MultiPolygon, Polygon, shape
from stactools.core.utils import antimeridian
from stactools.core.utils.antimeridian import Strategy

GSD_PATTERN: Final[Pattern[str]] = re.compile(r"[_R](\d0)m")


def extract_gsd(image_path: str) -> Optional[int]:
    match = GSD_PATTERN.search(image_path)
    if match:
        return int(match.group(1))
    else:
        return None


def fix_z_values(coord_values: List[str]) -> List[float]:
    """Some geometries have a '0' value in the z position
    of the coordinates. This method detects and removes z
    position coordinates. This assumes that in cases where
    the z value is included, it is included for all coordinates.
    """
    if len(coord_values) % 3 == 0:
        # Check if all 3rd position values are '0'
        # Ignore any blank values
        third_position_is_zero = [
            x == "0" for i, x in enumerate(coord_values) if i % 3 == 2 and x
        ]

        if all(third_position_is_zero):
            # Assuming that all 3rd position coordinates are z values
            # Remove them.
            return [float(c) for i, c in enumerate(coord_values) if i % 3 != 2]

    return [float(c) for c in coord_values if c]


def handle_antimeridian(item: Item, antimeridian_strategy: Strategy) -> None:
    """Handles some quirks of the antimeridian.
    Applies the requested SPLIT or NORMALIZE strategy via the stactools
    antimeridian utility. If the geometry is already SPLIT (a MultiPolygon,
    which can occur when using USGS geometry), a merged polygon with different
    longitude signs is created to match the expected input of the fix_item
    function.
    Args:
        item (Item): STAC Item
        antimeridian_strategy (Antimeridian): Either split on +/-180 or
            normalize geometries so all longitudes are either positive or
            negative.
    """
    geometry = shape(item.geometry)
    if isinstance(geometry, MultiPolygon):
        # force all positive lons so we can merge on an antimeridian split
        polys = list(geometry.geoms)
        for index, poly in enumerate(polys):
            coords = list(poly.exterior.coords)
            lons = [coord[0] for coord in coords]
            if min(lons) < 0:
                polys[index] = shapely.affinity.translate(poly, xoff=+360)
        merged_geometry = shapely.ops.unary_union(polys)

        # revert back to + and - lon signs for fix_item's expected input
        merged_coords = list(merged_geometry.exterior.coords)
        for index, coord in enumerate(merged_coords):
            if coord[0] > 180:
                merged_coords[index] = (coord[0] - 360, coord[1])
        item.geometry = Polygon(merged_coords)

    antimeridian.fix_item(item, antimeridian_strategy)



import json
from datetime import datetime
from typing import Any, Dict, Optional, Tuple

import pystac
from pystac.utils import str_to_datetime
from shapely.geometry import shape
from stactools.core.io import ReadHrefModifier, read_text

from stactools.sentinel2.constants import SENTINEL2_PROPERTY_PREFIX as s2_prefix
from stactools.sentinel2.constants import TILEINFO_METADATA_ASSET_KEY


class TileInfoMetadataError(Exception):
    pass


class TileInfoMetadata:
    def __init__(self, href, read_href_modifier: Optional[ReadHrefModifier] = None):
        self.href = href
        self.tileinfo = json.loads(read_text(self.href, read_href_modifier))

        self._datetime = str_to_datetime(self.tileinfo["timestamp"])
        self._geometry = self.tileinfo["tileDataGeometry"]
        self._bbox = shape(self._geometry).bounds
        self._product_path = self.tileinfo["productPath"]

    @property
    def product_path(self) -> str:
        return self._product_path

    @property
    def geometry(self) -> Dict[str, Any]:
        return self._geometry

    @property
    def bbox(self) -> Tuple[float, float, float, float]:
        return self._bbox

    @property
    def datetime(self) -> datetime:
        return self._datetime

    @property
    def metadata_dict(self):
        product_type = None
        product_name = self.tileinfo.get("productName")
        if product_name and "_MSIL2A_" in product_name:
            product_type = "S2MSI2A"
        elif product_name and "_MSIL1C_" in product_name:
            product_type = "S2MSI1C"

        result = {f"{s2_prefix}:product_type": product_type}

        return {k: v for k, v in result.items() if v is not None}

    def create_asset(self):
        asset = pystac.Asset(
            href=self.href, media_type=pystac.MediaType.JSON, roles=["metadata"]
        )
        return TILEINFO_METADATA_ASSET_KEY, asset
